﻿//
// ImageKeyUtils.cs
//
// Copyright (c) 2008 RednaxelaFX (Kris Mok, rednaxela0607@hotmail.com)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Reflection;
using System.Windows.Forms;
using ClassHierarchyViewer.Properties;

namespace ClassHierarchyViewer {
    internal static class ImageKeys {
        #region Image Key Constants

        internal static readonly string Namespace = "namespace";

        internal static readonly string ClassPublic = "class_public";
        internal static readonly string ClassPrivate = "class_private";
        internal static readonly string ClassProtected = "class_protected";
        internal static readonly string ClassInternal = "class_internal";

        internal static readonly string InterfacePublic = "interface_public";
        internal static readonly string InterfacePrivate = "interface_private";
        internal static readonly string InterfaceProtected = "interface_protected";
        internal static readonly string InterfaceInternal = "interface_internal";

        internal static readonly string DelegatePublic = "delegate_public";
        internal static readonly string DelegatePrivate = "delegate_private";
        internal static readonly string DelegateProtected = "delegate_protected";
        internal static readonly string DelegateInternal = "delegate_internal";

        internal static readonly string StructPublic = "struct_public";
        internal static readonly string StructPrivate = "struct_private";
        internal static readonly string StructProtected = "struct_protected";
        internal static readonly string StructInternal = "struct_internal";

        internal static readonly string EnumPublic = "enum_public";
        internal static readonly string EnumPrivate = "enum_private";
        internal static readonly string EnumProtected = "enum_protected";
        internal static readonly string EnumInternal = "enum_internal";

        internal static readonly string ConstructorPublic = "ctor_public";
        internal static readonly string ConstructorPrivate = "ctor_private";
        internal static readonly string ConstructorProtected = "ctor_protected";
        internal static readonly string ConstructorInternal = "ctor_internal";

        internal static readonly string MethodPublic = "method_public";
        internal static readonly string MethodPrivate = "method_private";
        internal static readonly string MethodProtected = "method_protected";
        internal static readonly string MethodInternal = "method_internal";

        internal static readonly string PropertyPublic = "property_public";
        internal static readonly string PropertyPrivate = "property_private";
        internal static readonly string PropertyProtected = "property_protected";
        internal static readonly string PropertyInternal = "property_internal";

        internal static readonly string FieldPublic = "field_public";
        internal static readonly string FieldPrivate = "field_private";
        internal static readonly string FieldProtected = "field_protected";
        internal static readonly string FieldInternal = "field_internal";

        internal static readonly string ConstantPublic = "constant_public";
        internal static readonly string ConstantPrivate = "constant_private";
        internal static readonly string ConstantProtected = "constant_protected";
        internal static readonly string ConstantInternal = "constant_internal";

        internal static readonly string EventPublic = "event";
        internal static readonly string EventPrivate = "event_private";
        internal static readonly string EventProtected = "event_protected";
        internal static readonly string EventInternal = "event_internal";

        #endregion
    }

    internal static class ImageKeyUtils {
        internal static void PopulateImageList( ImageList list ) {
            // namespace
            list.Images.Add( ImageKeys.Namespace, Resources.VSObject_Namespace );

            // class
            list.Images.Add( ImageKeys.ClassPublic, Resources.VSObject_Class );
            list.Images.Add( ImageKeys.ClassPrivate, Resources.VSObject_Class_Private );
            list.Images.Add( ImageKeys.ClassProtected, Resources.VSObject_Class_Protected );
            list.Images.Add( ImageKeys.ClassInternal, Resources.VSObject_Class_Internal );

            // interface
            list.Images.Add( ImageKeys.InterfacePublic, Resources.VSObject_Interface );
            list.Images.Add( ImageKeys.InterfacePrivate, Resources.VSObject_Interface_Private );
            list.Images.Add( ImageKeys.InterfaceProtected, Resources.VSObject_Interface_Protected );
            list.Images.Add( ImageKeys.InterfaceInternal, Resources.VSObject_Interface_Internal );

            // delegate
            list.Images.Add( ImageKeys.DelegatePublic, Resources.VSObject_Delegate );
            list.Images.Add( ImageKeys.DelegatePrivate, Resources.VSObject_Delegate_Private );
            list.Images.Add( ImageKeys.DelegateProtected, Resources.VSObject_Delegate_Protected );
            list.Images.Add( ImageKeys.DelegateInternal, Resources.VSObject_Delegate_Internal );

            // struct
            list.Images.Add( ImageKeys.StructPublic, Resources.VSObject_Structure );
            list.Images.Add( ImageKeys.StructPrivate, Resources.VSObject_Structure_Private );
            list.Images.Add( ImageKeys.StructProtected, Resources.VSObject_Structure_Protected );
            list.Images.Add( ImageKeys.StructInternal, Resources.VSObject_Structure_Internal );

            // enum
            list.Images.Add( ImageKeys.EnumPublic, Resources.VSObject_Enum );
            list.Images.Add( ImageKeys.EnumPrivate, Resources.VSObject_Enum_Private );
            list.Images.Add( ImageKeys.EnumProtected, Resources.VSObject_Enum_Protected );
            list.Images.Add( ImageKeys.EnumInternal, Resources.VSObject_Enum_Internal );

            // constructor
            list.Images.Add( ImageKeys.ConstructorPublic, Resources.VSObject_Constructor );
            list.Images.Add( ImageKeys.ConstructorProtected, Resources.VSObject_Constructor_Private );
            list.Images.Add( ImageKeys.ConstructorPrivate, Resources.VSObject_Constructor_Protected );
            list.Images.Add( ImageKeys.ConstructorInternal, Resources.VSObject_Constructor_Internal );

            // method
            list.Images.Add( ImageKeys.MethodPublic, Resources.VSObject_Method );
            list.Images.Add( ImageKeys.MethodPrivate, Resources.VSObject_Method_Private );
            list.Images.Add( ImageKeys.MethodProtected, Resources.VSObject_Method_Protected );
            list.Images.Add( ImageKeys.MethodInternal, Resources.VSObject_Method_Internal );

            // property
            list.Images.Add( ImageKeys.PropertyPublic, Resources.VSObject_Properties );
            list.Images.Add( ImageKeys.PropertyPrivate, Resources.VSObject_Properties_Private );
            list.Images.Add( ImageKeys.PropertyProtected, Resources.VSObject_Properties_Protected );
            list.Images.Add( ImageKeys.PropertyInternal, Resources.VSObject_Properties_Internal );

            // field
            list.Images.Add( ImageKeys.FieldPublic, Resources.VSObject_Field );
            list.Images.Add( ImageKeys.FieldPrivate, Resources.VSObject_Field_Private );
            list.Images.Add( ImageKeys.FieldProtected, Resources.VSObject_Field_Protected );
            list.Images.Add( ImageKeys.FieldInternal, Resources.VSObject_Field_Internal );

            // constant
            list.Images.Add( ImageKeys.ConstantPublic, Resources.VSObject_Constant );
            list.Images.Add( ImageKeys.ConstantPrivate, Resources.VSObject_Constant_Private );
            list.Images.Add( ImageKeys.ConstantProtected, Resources.VSObject_Constant_Protected );
            list.Images.Add( ImageKeys.ConstantInternal, Resources.VSObject_Constant_Internal );

            // event
            list.Images.Add( ImageKeys.EventPublic, Resources.VSObject_Event );
            list.Images.Add( ImageKeys.EventPrivate, Resources.VSObject_Event_Private );
            list.Images.Add( ImageKeys.EventProtected, Resources.VSObject_Event_Protected );
            list.Images.Add( ImageKeys.EventInternal, Resources.VSObject_Event_Internal );
        }

        private static Dictionary<string, string[ ]> _typeKeys = new Dictionary<string, string[ ]> {
            { "class", new [ ] {
                ImageKeys.ClassPublic,
                ImageKeys.ClassInternal,
                ImageKeys.ClassProtected,
                ImageKeys.ClassPrivate } },
            { "interface", new [ ] {
                ImageKeys.InterfacePublic,
                ImageKeys.InterfaceInternal,
                ImageKeys.InterfaceProtected,
                ImageKeys.InterfacePrivate } },
            { "delegate", new [ ] {
                ImageKeys.DelegatePublic,
                ImageKeys.DelegateInternal,
                ImageKeys.DelegateProtected,
                ImageKeys.DelegatePrivate } },
            { "struct", new [ ] {
                ImageKeys.StructPublic,
                ImageKeys.StructInternal,
                ImageKeys.StructProtected,
                ImageKeys.StructPrivate } },
            { "enum", new [ ] {
                ImageKeys.EnumPublic,
                ImageKeys.EnumInternal,
                ImageKeys.EnumProtected,
                ImageKeys.EnumPrivate } }
        };

        private static string GetTypeImageKeyCore( Type type, string[ ] keys ) {
            if ( type.IsPublic || type.IsNestedPublic ) {
                return keys[ 0 ]; // public
            } else if ( type.IsNotPublic || type.IsNestedAssembly ) {
                return keys[ 1 ]; // internal
            } else if ( type.IsNestedFamily || type.IsNestedFamANDAssem || type.IsNestedFamORAssem ) {
                return keys[ 2 ]; // protected
            } else {
                return keys[ 3 ]; // private
            }
        }

        internal static string GetTypeImageKey( Type type ) {
            if ( type.IsEnum ) {
                return GetTypeImageKeyCore( type, _typeKeys[ "enum" ] );
            } else if ( type.IsValueType ) {
                return GetTypeImageKeyCore( type, _typeKeys[ "struct" ] );
            } else if ( type.IsInterface ) {
                return GetTypeImageKeyCore( type, _typeKeys[ "interface" ] );
            } else if ( typeof( MulticastDelegate ) == type.BaseType ||
                ( typeof( Delegate ) == type.BaseType && typeof( MulticastDelegate ) != type ) ) {
                return GetTypeImageKeyCore( type, _typeKeys[ "delegate" ] );
            } else { // type.IsClass == true
                return GetTypeImageKeyCore( type, _typeKeys[ "class" ] );
            }
        }

        internal static string GetConstructorImageKey( ConstructorInfo ctor ) {
            if ( null == ctor ) {
                return string.Empty;
            } else if ( ctor.IsPublic ) {
                return ImageKeys.ConstructorPublic;
            } else if ( ctor.IsFamily ) {
                return ImageKeys.ConstructorInternal;
            } else if ( ctor.IsPrivate ) {
                return ImageKeys.ConstructorPrivate;
            } else { // protected
                return ImageKeys.ConstructorProtected;
            }
        }

        internal static string GetMethodImageKey( MethodInfo method ) {
            if ( null == method ) {
                return string.Empty;
            } else if ( method.IsPublic ) {
                return ImageKeys.MethodPublic;
            } else if ( method.IsAssembly ) {
                return ImageKeys.MethodInternal;
            } else if ( method.IsPrivate ) {
                return ImageKeys.MethodPrivate;
            } else { // protected
                return ImageKeys.MethodProtected;
            }
        }

        // TODO
        internal static string GetPropertyImageKey( PropertyInfo prop ) {
            var getter = prop.GetGetMethod( true );
            var setter = prop.GetSetMethod( true );

            Accessibility acc;
            if ( null == getter ) {
                acc = TypeUtils.GetMethodAccessibility( setter );
            } else if ( null == setter ) {
                acc = TypeUtils.GetMethodAccessibility( getter );
            } else {
                var accSetter = TypeUtils.GetMethodAccessibility( setter );
                var accGetter = TypeUtils.GetMethodAccessibility( getter );
                acc = accGetter > accSetter ? accGetter : accSetter;
            }

            switch ( acc ) {
            case Accessibility.Public:
                return ImageKeys.PropertyPublic;
            case Accessibility.Assembly:
                return ImageKeys.PropertyInternal;
            case Accessibility.Private:
                return ImageKeys.PropertyPrivate;
            default: // protected
                return ImageKeys.PropertyProtected;
            }
        }

        internal static string GetFieldImageKey( FieldInfo field ) {
            if ( field.IsInitOnly || field.IsLiteral ) {
                if ( field.IsPublic ) {
                    return ImageKeys.ConstantPublic;
                } else if ( field.IsAssembly ) {
                    return ImageKeys.ConstantInternal;
                } else if ( field.IsPrivate ) {
                    return ImageKeys.ConstantPrivate;
                } else { // protected
                    return ImageKeys.ConstantProtected;
                }
            } else {
                if ( field.IsPublic ) {
                    return ImageKeys.FieldPublic;
                } else if ( field.IsAssembly ) {
                    return ImageKeys.FieldInternal;
                } else if ( field.IsPrivate ) {
                    return ImageKeys.FieldPrivate;
                } else { // protected
                    return ImageKeys.FieldProtected;
                }
            }
        }

        internal static string GetEventImageKey( EventInfo evt ) {
            var addMethod = evt.GetAddMethod( true );
            var removeMethod = evt.GetRemoveMethod( true );

            Accessibility acc;
            if ( null == addMethod ) {
                acc = TypeUtils.GetMethodAccessibility( removeMethod );
            } else if ( null == removeMethod ) {
                acc = TypeUtils.GetMethodAccessibility( addMethod );
            } else {
                var accAdd = TypeUtils.GetMethodAccessibility( removeMethod );
                var accRemove = TypeUtils.GetMethodAccessibility( addMethod );
                acc = accRemove > accAdd ? accRemove : accAdd;
            }

            switch ( acc ) {
            case Accessibility.Public:
                return ImageKeys.EventPublic;
            case Accessibility.Assembly:
                return ImageKeys.EventInternal;
            case Accessibility.Private:
                return ImageKeys.EventPrivate;
            default: // protected
                return ImageKeys.EventProtected;
            }
        }

        internal static string GetMemberImageKey( MemberInfo member ) {
            var memberType = member.MemberType;
            switch ( memberType ) {
            case MemberTypes.Constructor:
                return GetConstructorImageKey( member as ConstructorInfo );
            case MemberTypes.Method:
                return GetMethodImageKey( member as MethodInfo );
            case MemberTypes.Property:
                return GetPropertyImageKey( member as PropertyInfo );
            case MemberTypes.Field:
                return GetFieldImageKey( member as FieldInfo );
            case MemberTypes.Event:
                return GetEventImageKey( member as EventInfo );
            }
            throw new Exception( "Should not reach here!" );
        }
    }
}
